// Copyright (C) 2022 - 2023, Xilinx Inc - All rights reserved
//
// Licensed under the Apache License, Version 2.0 (the "License"). You may
// not use this file except in compliance with the License. A copy of the
// License is located at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
// WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
// License for the specific language governing permissions and limitations
// under the License.

#include <memory>
#include <cstring>
#include "InternalDecoderSession.h"
#include "InternalDataBuffer.h"
#include "log_ama.h"

namespace xma {

#define XMA_DECODER_SESSION "DecoderSession"

/* The delay will be relative to the frame rate, so a framerate of 60 fps and
 * retry delay of 10 will lead to a delay of 1/(60*10) seconds */
#define XMA_DECODER_RETRY_DELAY 10
/* The retry count is relative to the retry delay (which is relative to the
 * framerate), so if the framerate is 60 fps and the retry count is
 * (XMA_DECODER_RETRY_DELAY*10), the maximum delay will be 10*1/60 or 1/6
 * seconds */
#define XMA_DECODER_MAX_RETRY_COUNT (XMA_DECODER_RETRY_DELAY * 10)

///////////////////////////////////////////////////////////////////////////////
InternalDecoderSession::InternalDecoderSession(IntSessionFactoryPtr sf, XmaDecoderProperties* props) : SessionBase(sf, "dec", props->params, props->param_cnt) {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::InternalDecoderSession()");

  snprintf(m_instance_name, 64, "decoder_@%p", this);
  LogAmaInstanceMgmt ama_log_env(LOG_AMA_ACCEL_DEC, m_instance_name);

  m_mem_ctx = vpi_mem_handle_t(m_sf->CreateMemHandle(), [](VpiMemHandle mem_ctx) { vpi_memory_release(mem_ctx); });
  if (!m_mem_ctx) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::InternalDecoderSession() Failed to create VPI memory!");
  }

  memset(&m_params, 0, sizeof(VpiDecInitParams));
  VpiHwId hw_id;
  switch (props->hwdecoder_type) {
  case XMA_H264_DECODER_TYPE:
    hw_id             = VPE_HW_H264DEC;
    m_params.dec_type = VPE_DEC_H264;
    break;
  case XMA_HEVC_DECODER_TYPE:
    hw_id             = VPE_HW_HEVCDEC;
    m_params.dec_type = VPE_DEC_HEVC;
    break;
  case XMA_JPEG_DECODER_TYPE:
    hw_id             = VPE_HW_JPEGDEC;
    m_params.dec_type = VPE_DEC_JPEG;
    break;
  case XMA_VP9_DECODER_TYPE:
    hw_id             = VPE_HW_VP9DEC;
    m_params.dec_type = VPE_DEC_VP9;
    break;
  case XMA_AV1_DECODER_TYPE:
    hw_id             = VPE_HW_AV1DEC;
    m_params.dec_type = VPE_DEC_AV1;
    break;
  default:
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::InternalDecoderSession() Invalid decoder type!");
  }
  m_params.hw_ctx     = m_sf->GetContext();
  m_params.mem_ctx    = m_mem_ctx.get();
  m_params.src_width  = props->width;
  m_params.src_height = props->height;
  m_params.resize_str = nullptr;
  m_params.src_depth  = props->bits_per_pixel;

  VpiHwCfg* hw_cfg = vpi_device_get_hw_config(m_sf->GetDevHandle(), hw_id);

  if (hw_cfg->sw_formats[0] == VPI_FORMAT_NONE) {
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::InternalDecoderSession() Hardware error format!");
  }

  if (hw_cfg->hw_formats[0] != VPI_FORMAT_VPE) {
    THROW_INVALID_ARGUMENT_EXCEPTION(
        m_log, XMA_ERROR_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::InternalDecoderSession() Hardware does not offer any usable surface formats!");
  }
  vpi_device_free_hw_config(hw_cfg);

  m_params.transcode = 0;

  VpiBufferPool* new_pool = nullptr;
  VpiRet         ret      = vpi_buffer_pool_create(m_sf->GetMemHandle(), XMA_DEFAULT_PKT_POOL_SIZE, XMA_DEFAULT_PKT_SIZE, RC_SIDE_EN, false, 0, &new_pool);
  if (ret < VPI_SUCCESS) {
    THROW_RUNTIME_EXCEPTION(
        m_log, XMA_ERROR_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::InternalDecoderSession() Failed to create buffer pool: %s!", vpi_error_str(ret));
  }
  m_pool            = pool_handle_t(new_pool, [&](void* pool) {
    if (pool) {
      if (!m_resetting) {
        DataBuffer* buf = static_cast<DataBuffer*>(vpi_decoder_flush_pkt(pool));
        while (buf) {
          lock_t guard(m_mutex);
          m_buffer_map[buf].second->DecRef();
          m_buffer_map.erase(buf);
          guard.unlock();
          buf = static_cast<DataBuffer*>(vpi_decoder_flush_pkt(pool));
        }
      }
      vpi_pool_unref(pool);
    }
  });
  m_params.pkt_pool = m_pool.get();

  XmaParameter* param;
  if ((param = GetParameter(props->params, props->param_cnt, XMA_DEC_PARAM_LOW_LATENCY)) != nullptr) {
    if (*(int32_t*) param->value < 0 || *(int32_t*) param->value > 1) {
      THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::InternalDecoderSession() Invalid value for low_latency");
    } else {
      m_low_latency = static_cast<bool>(*(int32_t*) param->value);
    }
  } else {
    m_low_latency = XMA_DEC_LOW_LATENCY_DEFAULT;
  }
  m_params.low_latency = m_low_latency;

  const char* fmt = nullptr;
  param           = GetParameter(props->params, props->param_cnt, XMA_DEC_OUTPUT_FORMAT);
  if (param) {
    switch (*static_cast<int32_t*>(param->value)) {
    case XMA_NV12_FMT_TYPE:
      if (props->bits_per_pixel != 8) {
        THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_DECODER_SESSION,
            "InternalDecoderSession::InternalDecoderSession() Input bit depth (%d) and output bit depth (8) do not match!", props->bits_per_pixel);
      }
      if (props->flags & XMA_FRAME_PROPERTY_FLAG_TILE_4x4) {
        if (props->flags & XMA_FRAME_PROPERTY_FLAG_COMPRESS) {
          fmt = "nv12-ma";
        } else {
          fmt = "nv12-tile";
        }
      } else {
        if (props->flags & XMA_FRAME_PROPERTY_FLAG_COMPRESS) {
          THROW_INVALID_ARGUMENT_EXCEPTION(
              m_log, XMA_ERROR_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::InternalDecoderSession() NV12 pixel format does not support compression without tile!");
        } else {
          fmt = "nv12";
        }
      }
      break;
    case XMA_P010LE_FMT_TYPE:
      if (props->bits_per_pixel != 10) {
        THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_DECODER_SESSION,
            "InternalDecoderSession::InternalDecoderSession() Input bit depth (%d) and output bit depth (10) do not match!", props->bits_per_pixel);
      }
      if (props->flags & XMA_FRAME_PROPERTY_FLAG_TILE_4x4) {
        if (props->flags & XMA_FRAME_PROPERTY_FLAG_COMPRESS) {
          fmt = "p010le-ma";
        } else {
          fmt = "p010le-tile";
        }
      } else {
        if (props->flags & XMA_FRAME_PROPERTY_FLAG_COMPRESS) {
          THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_DECODER_SESSION,
              "InternalDecoderSession::InternalDecoderSession() P010LE pixel format does not support compression without tile!");
        } else {
          fmt = "p010le";
        }
      }
      break;
    case XMA_PACKED10_FMT_TYPE:
      if (props->bits_per_pixel != 10) {
        THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_DECODER_SESSION,
            "InternalDecoderSession::InternalDecoderSession() Input bit depth (%d) and output bit depth (10) do not match!", props->bits_per_pixel);
      }
      if (props->flags & (XMA_FRAME_PROPERTY_FLAG_TILE_4x4 | XMA_FRAME_PROPERTY_FLAG_COMPRESS)) {
        THROW_INVALID_ARGUMENT_EXCEPTION(
            m_log, XMA_ERROR_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::InternalDecoderSession() Packed10 pixel format does not support compression or tile!");
      }
      fmt = "packed10";
      break;
    case XMA_RGB24_P_FMT_TYPE:
      if (props->flags & (XMA_FRAME_PROPERTY_FLAG_TILE_4x4 | XMA_FRAME_PROPERTY_FLAG_COMPRESS)) {
        THROW_INVALID_ARGUMENT_EXCEPTION(
            m_log, XMA_ERROR_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::InternalDecoderSession() RGBP pixel format does not support compression or tile!");
      }
      fmt = "rgbp";
      break;
    case XMA_YUV420P_FMT_TYPE:
      if (props->bits_per_pixel != 8) {
        THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_DECODER_SESSION,
            "InternalDecoderSession::InternalDecoderSession() Input bit depth (%d) and output bit depth (8) do not match!", props->bits_per_pixel);
      }
      if (props->flags & (XMA_FRAME_PROPERTY_FLAG_TILE_4x4 | XMA_FRAME_PROPERTY_FLAG_COMPRESS)) {
        THROW_INVALID_ARGUMENT_EXCEPTION(
            m_log, XMA_ERROR_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::InternalDecoderSession() YUV420P pixel format does not support compression or tile!");
      }
      fmt = "yuv420p";
      break;
    case XMA_YUV420P10LE_FMT_TYPE:
      if (props->bits_per_pixel != 10) {
        THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_DECODER_SESSION,
            "InternalDecoderSession::InternalDecoderSession() Input bit depth (%d) and output bit depth (10) do not match!", props->bits_per_pixel);
      }
      if (props->flags & (XMA_FRAME_PROPERTY_FLAG_TILE_4x4 | XMA_FRAME_PROPERTY_FLAG_COMPRESS)) {
        THROW_INVALID_ARGUMENT_EXCEPTION(
            m_log, XMA_ERROR_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::InternalDecoderSession() YUV420P10LE pixel format does not support compression or tile!");
      }
      fmt = "yuv420p10le";
      break;
    default:
      THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::InternalDecoderSession() Unsupported output pixel format type!");
    }
  } else {
    if (props->bits_per_pixel == 8) {
      fmt = "nv12-ma";
    } else if (props->bits_per_pixel == 10) {
      fmt = "p010le-ma";
    } else {
      THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::InternalDecoderSession() Unsupported bits per pixel!");
    }
  }

  m_params.outfmt_str = const_cast<char*>(fmt);

  lock_t guard(m_mutex);

  m_handle = dec_handle_t(vpi_decoder_init(&m_params), [](VpiDecHandle handle) { vpi_decoder_close(handle); });
  if (!m_handle.get()) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::InternalDecoderSession() Failed to initialize decoder!");
  }

  param = GetParameter(props->params, props->param_cnt, "enable_mc");
  if (param) {
    char text[64];
    sprintf(text, "%d", *static_cast<int32_t*>(param->value));
    if (vpi_decoder_set_param(m_handle.get(), const_cast<char*>("enable_mc"), text) < 0) {
      m_handle.reset();
      THROW_RUNTIME_EXCEPTION(
          m_log, XMA_ERROR_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::InternalDecoderSession() Unable to set decoder parameter enable_mc = \"%s\"", text);
    }
  }

  if (m_sf->GetAPIVersion() >= XMA_API_VERSION_1_1) {
    param = GetParameter(props->params, props->param_cnt, XMA_DEC_PARAM_THREADS);
    if (param) {
      m_use_thread = ((*static_cast<int32_t*>(param->value)) ? 1 : 0);
      if (m_use_thread) {
        param = GetParameter(props->params, props->param_cnt, XMA_DEC_PARAM_WAIT);
        if (param) {
          m_wait = *static_cast<int32_t*>(param->value);
        }
      }
    }

    param = GetParameter(props->params, props->param_cnt, XMA_DEC_PARAM_PROP_CHANGE_CALLBACK);
    if (param) {
      m_prop_callback = reinterpret_cast<xma::DecPropChangeCallbackFunction>(param->value);
      param           = GetParameter(props->params, props->param_cnt, XMA_DEC_PARAM_PROP_CHANGE_PARAM);
      if (param) {
        m_prop_param = param->value;
      }
    }
  }

  if (vpi_decoder_open(m_handle.get()) < 0) {
    m_handle.reset();
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::InternalDecoderSession() Unable to open decoder for processing!");
  }

  m_pkt = pkt_handle_t(static_cast<VpiPkt*>(vpi_memory_get_struct(VPE_PACKET)), [](VpiPkt* pkt) { vpi_memory_free_struct(pkt); });

  VpiFrm* pic = nullptr;
  if ((vpi_frame_alloc(m_sf->GetMemHandle(), &pic) < 0) || !pic) {
    m_handle.reset();
    THROW_BAD_ALLOC_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::InternalDecoderSession() Out of memory!");
  }
  InitVpiFrm(pic);
  m_pic       = pic_handle_t(pic, [&](VpiFrm* _pic) { vpi_frame_free(m_sf->GetMemHandle(), _pic); });
  m_framerate = props->framerate;

  if (m_use_thread) {
    m_thread = std::thread(InternalDecoderSession::Thread, this);
  }

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::InternalDecoderSession() - Returning");
}

///////////////////////////////////////////////////////////////////////////////
InternalDecoderSession::~InternalDecoderSession() {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::~InternalDecoderSession()");

  if (!m_eos_pushed) {
    m_log->LogMsg(XMA_WARNING_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::~InternalDecoderSession() - End of stream not sent!");
  } else if (!m_eos_pulled) {
    m_log->LogMsg(XMA_WARNING_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::~InternalDecoderSession() - End of stream not recieved!");
  }

  Flush();

  m_handle.reset();
  m_pool.reset();
}

///////////////////////////////////////////////////////////////////////////////
FramePropertiesPtr InternalDecoderSession::GetProps() const {
  lock_t guard(m_mutex);

  return m_props;
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalDecoderSession::SetLog(LogPtr log) {
  lock_t guard(m_mutex);
  SessionBase::SetLog(log);
  return XMA_SUCCESS;
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalDecoderSession::Enqueue(queue_t& data) {
  LogAmaInstanceMgmt ama_log_env(LOG_AMA_ACCEL_DEC, m_instance_name);
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::Enqueue()");
  DataBufferPtr    buf = std::move(std::get<QueueDataBuffer>(data).data);
  IntDataBufferPtr int_buf;
  if (buf) {
    LogLatencySend(m_frame_enqueued + 1);
    int_buf = std::dynamic_pointer_cast<InternalDataBuffer>(buf);
    if (int_buf) {
      int_buf->IncRef();
    } else {
      THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::Enqueue() invalid data buffer!");
    }
  } else if (m_eos_pulled) {
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::Enqueue() - Returning success");
    return XMA_SUCCESS;
  }

  lock_t guard(m_mutex);

  if (m_use_thread) {
    if (m_buf || m_resetting) {
      if (m_wait) {
        do {
          m_wait_enqueue_cond.wait(guard);
          if (m_exception.HasException()) {
            m_exception.RethrowException();
          }
          if (m_stop) {
            m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::Enqueue() - Returning XMA_ERROR_TIMEOUT: time alloted for call exceeded");
            return XMA_ERROR_TIMEOUT;
          }
        } while (m_buf || m_resetting);
      } else {
        if (int_buf) {
          int_buf->DecRef();
        }
        m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::Enqueue() - Returning try again");
        return XMA_TRY_AGAIN;
      }
    }
    if (!buf) {
      m_eos_pushed = true;
    } else {
      m_buf = std::move(std::dynamic_pointer_cast<InternalDataBuffer>(buf));
    }
    m_thread_enqueue_cond.notify_one();

    ++m_frame_enqueued;
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::Enqueue() - Returning success");
    return XMA_SUCCESS;
  }

  if (!m_handle.get()) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::Enqueue() Decoder not initialized!");
  }

  if (m_frame) {
    if (int_buf) {
      int_buf->DecRef();
    }
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::Enqueue() - Returning try again");
    return XMA_TRY_AGAIN;
  }

  int32_t ret = PushPacket(buf);
  if (ret != XMA_SUCCESS) {
    char ret_str[64];
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::Enqueue() - Returning %s", xma_perror(ret, ret_str, sizeof(ret_str)));
    return ret;
  }

  m_frame = GetFrame();
  if (m_frame) {
    if (!m_props) {
      m_props = m_frame->GetProps();
    }
  } else if (!m_eos_found) {
    if (!buf) {
      m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::Enqueue() - Returning try again");
      return XMA_TRY_AGAIN;
    } else if (!m_props) {
      ++m_frame_enqueued;
      m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::Enqueue() - Returning send more data");
      return XMA_SEND_MORE_DATA;
    }
  }

  ++m_frame_enqueued;
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::Enqueue() - Returning success");
  return XMA_SUCCESS;
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalDecoderSession::Dequeue(queue_t& data) {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::Dequeue()");
  data = QueueFrame();

  if (m_eos_pulled) {
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::Dequeue() - Returning EOS");
    return XMA_EOS;
  }

  lock_t guard(m_mutex);

  FramePtr frame;
  if (m_use_thread) {
    if (m_wait) {
      while (((!m_frame) && (!m_eos_found)) || m_resetting) {
        m_wait_dequeue_cond.wait(guard);
        if (m_exception.HasException()) {
          m_exception.RethrowException();
        }
        if (m_stop) {
          m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::Dequeue() - Returning XMA_ERROR_TIMEOUT: time alloted for call exceeded");
          return XMA_ERROR_TIMEOUT;
        }
      }
    } else if ((!m_frame) && (!m_eos_found)) {
      m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::Dequeue() - Returning resend and receive");
      return XMA_RESEND_AND_RECV;
    }
    frame = std::move(m_frame);
    m_thread_dequeue_cond.notify_all();
  } else if (m_frame) {
    frame = std::move(m_frame);
  } else {
    frame = GetFrame();
  }

  std::get<QueueFrame>(data).frame = frame;

  if (!frame) {
    if (m_eos_found) {
      m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::Dequeue() - Returning EOS");
      m_eos_pulled = true;
      return XMA_EOS;
    } else {
      m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::Dequeue() - Returning resend and receive");
      return XMA_RESEND_AND_RECV;
    }
  }

  FramePropertiesPtr new_props = frame->GetProps();
  if (!m_props) {
    m_props = new_props;
  }
  if ((new_props->GetWidth() != m_props->GetWidth()) || (new_props->GetHeight() != m_props->GetHeight())) {
    m_log->LogMsg(XMA_INFO_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::Dequeue() New resolution detected (%dx%d).", new_props->GetWidth(), new_props->GetHeight());
    if (m_prop_callback) {
      int32_t ret = m_prop_callback(new_props, m_prop_param);
      if (ret == XMA_SUCCESS) {
        m_props = new_props;
        LogLatencyRecv(++m_frame_dequeued);
        m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::Dequeue() - Returning success");
        return XMA_SUCCESS;
      } else if (ret == XMA_EOS) {
        m_log->LogMsg(XMA_INFO_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::Dequeue() App forcing EOS");
        Flush();
        m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::Dequeue() - Returning forced EOS");
        return XMA_EOS;
      } else {
        Flush();
        char ret_str[64];
        m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::Dequeue() - Returning %s", xma_perror(ret, ret_str, sizeof(ret_str)));
        return ret;
      }
    } else {
      Flush();
      THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::Dequeue() Unsupported properties change!");
    }
  }

  LogLatencyRecv(++m_frame_dequeued);
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::Dequeue() - Returning success");
  return XMA_SUCCESS;
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalDecoderSession::PushPacket(DataBufferPtr buf) {
  LogAmaInstanceMgmt ama_log_env(LOG_AMA_ACCEL_DEC, m_instance_name);
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::PushPacket()");

  bool wait = false;
  if (buf) {
    m_pkt->buffer           = buf->GetBuffer();
    m_pkt->size             = buf->GetSize();
    m_pkt->pts              = XMA_AV_NOPTS_VALUE;
    m_pkt->dts              = m_frame_sent;
    m_pkt->opaque           = buf.get();
    m_buffer_map[buf.get()] = std::make_pair(m_frame_sent, std::dynamic_pointer_cast<InternalDataBuffer>(buf));
    m_pts_map[m_frame_sent] = buf->GetPts();
    buf.reset();
  } else {
    m_eos_pushed  = true;
    wait          = true;
    m_pkt->size   = 0;
    m_pkt->pts    = XMA_AV_NOPTS_VALUE;
    m_pkt->dts    = XMA_AV_NOPTS_VALUE;
    m_pkt->opaque = nullptr;
  }

  int32_t ret = 1;
  if (wait || (m_pkt->size > 0)) {
    ret = vpi_decoder_put_packet(m_handle.get(), m_pkt.get());
    if (ret < 0) {
      THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::PushPacket() Failed to send data to decoder!");
    }
    m_pkt->size -= ret;
  }

  if ((!wait) && (ret == 0)) {
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::PushPacket() - Returning try again");
    return XMA_TRY_AGAIN;
  }

  if (!wait) {
    ++m_frame_sent;
  }

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::PushPacket() - Returning success");
  return XMA_SUCCESS;
}

///////////////////////////////////////////////////////////////////////////////
IntFramePtr InternalDecoderSession::GetFrame() {
  LogAmaInstanceMgmt ama_log_env(LOG_AMA_ACCEL_DEC, m_instance_name);
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::GetFrame()");
  DataBuffer* buf = static_cast<DataBuffer*>(vpi_decoder_release_pkt(m_handle.get()));
  while (buf) {
    m_buffer_map[buf].second->DecRef();
    m_buffer_map.erase(buf);
    buf = static_cast<DataBuffer*>(vpi_decoder_release_pkt(m_handle.get()));
  }

  int32_t ret = vpi_decoder_get_frame(m_handle.get());
  if (ret == VPI_DEC_FRAME) {
    ret = vpi_decoder_get_output_frame(m_handle.get(), m_pic.get());
    if (ret) {
      THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::GetFrame() Failed to decode frame!");
    }

    m_pic->pts = m_pts_map[m_pic->dts];
    m_pts_map.erase(m_pic->dts);
    IntFramePtr frame = std::make_shared<InternalFrame>(m_sf, m_log, m_pic.release(), m_framerate);

    VpiBufferPool* vpi_bp    = frame->GetBufferPool();
    int32_t        pool_size = vpi_buffer_pool_get_depth(vpi_bp);
    if (pool_size > m_pool_size) {
      m_pool_size = pool_size;
    } else if (pool_size < m_pool_size) {
      int32_t extend_size = m_pool_size - pool_size;
      if (extend_size > 0) {
        frame->ExtendBufferPool(extend_size);
      }
    }

    VpiFrm* pic = nullptr;
    if ((vpi_frame_alloc(m_sf->GetMemHandle(), &pic) < 0) || !pic) {
      THROW_BAD_ALLOC_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::GetFrame() Out of memory!");
    }
    InitVpiFrm(pic);
    m_pic = pic_handle_t(pic, [&](VpiFrm* _pic) { vpi_frame_free(m_sf->GetMemHandle(), _pic); });

    m_frame_recv++;
    frame->SetFrozen(true);

    m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::GetFrame() - Returning frame");
    return frame;
  } else if ((ret == VPI_DEC_RES_CHANGED_LOWER) || (ret == VPI_DEC_RES_CHANGED_HIGHER)) {
    int64_t start_frame = m_frame_recv;
    int64_t end_frame   = m_frame_sent;
    m_resetting         = true;
    ret                 = vpi_decoder_get_output_frame(m_handle.get(), m_pic.get());
    if (ret) {
      THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::GetFrame() Failed to decode frame!");
    }
    int32_t pool_size = vpi_buffer_pool_get_depth(reinterpret_cast<VpiBufferPool*>(m_pool.get()));
    m_handle.reset();
    m_pool.reset();
    m_params.src_width  = m_pic->width;
    m_params.src_height = m_pic->height;

    VpiBufferPool* new_pool = nullptr;
    ret                     = vpi_buffer_pool_create(m_sf->GetMemHandle(), pool_size, XMA_DEFAULT_PKT_SIZE, RC_SIDE_EN, false, 0, &new_pool);
    if (ret < VPI_SUCCESS) {
      THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::GetFrame() Failed to create buffer pool: %s!", vpi_error_str(ret));
    }
    m_pool            = pool_handle_t(new_pool, [&](void* pool) {
      if (pool) {
        if (!m_resetting) {
          void* p = vpi_decoder_flush_pkt(pool);
          while (p) {
            lock_t guard(m_mutex);
            buf = reinterpret_cast<DataBuffer*>(p);
            m_buffer_map[buf].second->DecRef();
            m_buffer_map.erase(buf);
            guard.unlock();
            p = vpi_decoder_flush_pkt(pool);
          }
        }
        vpi_pool_unref(pool);
      }
    });
    m_params.pkt_pool = m_pool.get();

    m_handle = dec_handle_t(vpi_decoder_init(&m_params), [](VpiDecHandle handle) { vpi_decoder_close(handle); });
    if (!m_handle.get()) {
      THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::GetFrame() Failed to initialize decoder!");
    }

    if (vpi_decoder_open(m_handle.get()) < 0) {
      m_handle.reset();
      THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::GetFrame() Unable to open decoder for processing!");
    }

    m_frame_sent = m_frame_recv;
    for (; start_frame < end_frame; ++start_frame) {
      for (auto it = m_buffer_map.begin(); it != m_buffer_map.end(); ++it) {
        if (it->second.first == start_frame) {
          do {
            ret = PushPacket(it->second.second);
            if (ret == XMA_TRY_AGAIN) {
              usleep(100);
            }
          } while (ret == XMA_TRY_AGAIN);
          break;
        }
      }
    }

    InitVpiFrm(m_pic.get());

    m_wait_enqueue_cond.notify_all();
    m_wait_dequeue_cond.notify_all();
    m_resetting       = false;
    IntFramePtr frame = GetFrame();
    if (frame) {
      m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::GetFrame() - Returning frame");
      return frame;
    } else if (m_eos_found) {
      m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::GetFrame() - Returning EOS");
      return nullptr;
    } else {
      m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::GetFrame() - Returning NULL");
      return nullptr;
    }
  } else if (ret == VPI_DEC_EOF) {
    m_eos_found = true;
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::GetFrame() - Returning EOS");
    return nullptr;
  } else if (ret == -1) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::GetFrame() Failed to get frame!");
  } else {
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::GetFrame() - Returning NULL");
    return nullptr;
  }
}

///////////////////////////////////////////////////////////////////////////////
void InternalDecoderSession::Flush() {
  LogAmaInstanceMgmt ama_log_env(LOG_AMA_ACCEL_DEC, m_instance_name);
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::Flush()");

  if (m_use_thread) {
    m_stop = true;
    m_thread_enqueue_cond.notify_all();
    m_thread_dequeue_cond.notify_all();
    m_wait_enqueue_cond.notify_all();
    m_wait_dequeue_cond.notify_all();
    m_thread.join();
  }

  while (!m_eos_found) {
    PushPacket(nullptr);
    GetFrame();
  }
  m_eos_pulled = true;

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::Flush() - Returning");
}

///////////////////////////////////////////////////////////////////////////////
void InternalDecoderSession::Thread(InternalDecoderSession* session) {
  LogAmaInstanceMgmt ama_log_env(LOG_AMA_ACCEL_DEC, session->m_instance_name);
  session->m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::Thread()");

  while ((!session->m_stop) && (!session->m_eos_found)) {
    lock_t lock(session->m_mutex);
    while (((!session->m_buf) && (!session->m_eos_pushed)) || session->m_resetting) {
      session->m_thread_enqueue_cond.wait(lock);
      if (session->m_stop) {
        return;
      }
    }

    IntDataBufferPtr buf = std::move(session->m_buf);
    session->m_wait_enqueue_cond.notify_all();

    try {
      session->PushPacket(buf);
      IntFramePtr frame = session->GetFrame();

      if (session->m_stop) {
        return;
      }
      if (frame) {
        while (session->m_frame) {
          session->m_thread_dequeue_cond.wait(lock);
          if (session->m_stop) {
            return;
          }
        }
        session->m_frame = std::move(frame);
        session->m_wait_dequeue_cond.notify_one();
      } else if (session->m_eos_found) {
        while (session->m_frame) {
          session->m_thread_dequeue_cond.wait(lock);
          if (session->m_stop) {
            return;
          }
        }
        session->m_wait_dequeue_cond.notify_one();
        return;
      }
    } catch (ThreadThrowable& exception) {
      session->m_exception.SetException(exception.CreateClone());
      session->m_stop = true;
      session->m_wait_enqueue_cond.notify_all();
      session->m_wait_dequeue_cond.notify_all();
      return;
    } catch (...) {
      session->m_exception.SetException(std::make_shared<UnknownException>());
      session->m_stop = true;
      session->m_wait_enqueue_cond.notify_all();
      session->m_wait_dequeue_cond.notify_all();
      return;
    }
  }
}

} // namespace xma
